@@ -5,7 +5,7 @@
  * object level logging, and fully-qualified object names for all DML and DDL
  * statements where possible (See README.md for details).
  *
- * Copyright (c) 2014-2015, PostgreSQL Global Development Group
+ * Copyright (c) 2014-2020, PostgreSQL Global Development Group
  *------------------------------------------------------------------------------
  */
 #include "postgres.h"
@@ -25,6 +25,7 @@
 #include "miscadmin.h"
 #include "libpq/auth.h"
 #include "nodes/nodes.h"
+#include "nodes/params.h"
 #include "tcop/utility.h"
 #include "tcop/deparse_utility.h"
 #include "utils/acl.h"
@@ -494,7 +495,7 @@ log_audit_event(AuditEventStackItem *stackItem)
 
             switch (stackItem->auditEvent.commandTag)
             {
-                    /* Currently, only EXECUTE is different */
+                /* Currently, only EXECUTE is different */
                 case T_ExecuteStmt:
                     className = CLASS_MISC;
                     class = LOG_MISC;
@@ -596,7 +597,7 @@ log_audit_event(AuditEventStackItem *stackItem)
         case LOGSTMT_ALL:
             switch (stackItem->auditEvent.commandTag)
             {
-                    /* READ statements */
+                /* READ statements */
                 case T_CopyStmt:
                 case T_SelectStmt:
                 case T_PrepareStmt:
@@ -605,7 +606,7 @@ log_audit_event(AuditEventStackItem *stackItem)
                     class = LOG_READ;
                     break;
 
-                    /* FUNCTION statements */
+                /* FUNCTION statements */
                 case T_DoStmt:
                     className = CLASS_FUNCTION;
                     class = LOG_FUNCTION;
@@ -969,7 +970,7 @@ log_select_dml(Oid auditOid, List *rangeTabls)
     foreach(lr, rangeTabls)
     {
         Oid relOid;
-        Relation rel;
+        Oid relNamespaceOid;
         RangeTblEntry *rte = lfirst(lr);
 
         /* We only care about tables, and can ignore subqueries etc. */
@@ -992,13 +993,10 @@ log_select_dml(Oid auditOid, List *rangeTabls)
          * false) then filter out any system relations here.
          */
         relOid = rte->relid;
-        rel = relation_open(relOid, NoLock);
+        relNamespaceOid = get_rel_namespace(relOid);
 
-        if (!auditLogCatalog && IsSystemNamespace(RelationGetNamespace(rel)))
-        {
-            relation_close(rel, NoLock);
+        if (!auditLogCatalog && IsSystemNamespace(relNamespaceOid))
             continue;
-        }
 
         /*
          * Default is that this was not through a grant, to support session
@@ -1095,10 +1093,8 @@ log_select_dml(Oid auditOid, List *rangeTabls)
 
         /* Get a copy of the relation name and assign it to object name */
         auditEventStack->auditEvent.objectName =
-            quote_qualified_identifier(get_namespace_name(
-                                           RelationGetNamespace(rel)),
-                                       RelationGetRelationName(rel));
-        relation_close(rel, NoLock);
+            quote_qualified_identifier(
+                get_namespace_name(relNamespaceOid), get_rel_name(relOid));
 
         /* Perform object auditing only if the audit role is valid */
         if (auditOid != InvalidOid)
@@ -1293,7 +1289,7 @@ pgaudit_ExecutorStart_hook(QueryDesc *queryDesc, int eflags)
 
         /* Initialize the audit event */
         stackItem->auditEvent.commandText = queryDesc->sourceText;
-        stackItem->auditEvent.paramList = queryDesc->params;
+        stackItem->auditEvent.paramList = copyParamList(queryDesc->params);
     }
 
     /* Call the previous hook or standard function */
@@ -1343,11 +1339,11 @@ pgaudit_ExecutorCheckPerms_hook(List *rangeTabls, bool abort)
  */
 static void
 pgaudit_ProcessUtility_hook(Node *parsetree,
-                             const char *queryString,
-                             ProcessUtilityContext context,
-                             ParamListInfo params,
-                             DestReceiver *dest,
-                             char *completionTag)
+                            const char *queryString,
+                            ProcessUtilityContext context,
+                            ParamListInfo params,
+                            DestReceiver *dest,
+                            char *completionTag)
 {
     AuditEventStackItem *stackItem = NULL;
     int64 stackId = 0;
@@ -1361,11 +1357,30 @@ pgaudit_ProcessUtility_hook(Node *parsetree,
         /* Process top level utility statement */
         if (context == PROCESS_UTILITY_TOPLEVEL)
         {
+            /*
+             * If the stack is not empty then the only allowed entries are open
+             * select, show, and explain cursors
+             */
             if (auditEventStack != NULL)
-                elog(ERROR, "pgaudit stack is not empty");
+            {
+                AuditEventStackItem *nextItem = auditEventStack;
+
+                do
+                {
+                    if (nextItem->auditEvent.commandTag != T_SelectStmt &&
+                        nextItem->auditEvent.commandTag != T_VariableShowStmt &&
+                        nextItem->auditEvent.commandTag != T_ExplainStmt)
+                    {
+                        elog(ERROR, "pgaudit stack is not empty");
+                    }
+
+                    nextItem = nextItem->next;
+                }
+                while (nextItem != NULL);
+            }
 
             stackItem = stack_push();
-            stackItem->auditEvent.paramList = params;
+            stackItem->auditEvent.paramList = copyParamList(params);
         }
         else
             stackItem = stack_push();
@@ -1384,6 +1399,19 @@ pgaudit_ProcessUtility_hook(Node *parsetree,
             stackItem->auditEvent.commandTag == T_DoStmt &&
             !IsAbortedTransactionBlockState())
             log_audit_event(stackItem);
+
+        /*
+         * A close will free the open cursor which will also free the close
+         * audit entry. Immediately log the close and set stackItem to NULL so
+         * it won't be logged later.
+         */
+        if (stackItem->auditEvent.commandTag == T_ClosePortalStmt)
+        {
+            if (auditLogBitmap & LOG_MISC && !IsAbortedTransactionBlockState())
+                log_audit_event(stackItem);
+
+            stackItem = NULL;
+        }
     }
 
     /* Call the standard process utility chain. */
@@ -1816,7 +1844,7 @@ _PG_init(void)
     if (inited)
         return;
 
-    /* Must be loaded with shared_preload_libaries */
+    /* Must be loaded with shared_preload_libraries */
     if (!process_shared_preload_libraries_in_progress)
         ereport(ERROR, (errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
                 errmsg("pgaudit must be loaded via shared_preload_libraries")));
@@ -1845,7 +1873,7 @@ _PG_init(void)
 
         "Specifies that session logging should be enabled in the case where "
         "all relations in a statement are in pg_catalog.  Disabling this "
-         "setting will reduce noise in the log from tools like psql and PgAdmin "
+        "setting will reduce noise in the log from tools like psql and PgAdmin "
         "that query the catalog heavily.",
 
         NULL,
@@ -1972,7 +2000,11 @@ _PG_init(void)
     object_access_hook = pgaudit_object_access_hook;
 
     /* Log that the extension has completed initialization */
+#ifndef EXEC_BACKEND
     ereport(LOG, (errmsg("pgaudit extension initialized")));
+#else
+    ereport(DEBUG1, (errmsg("pgaudit extension initialized")));
+#endif /* EXEC_BACKEND */
 
     inited = true;
 }
